#!/usr/bin/env python
# License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>


BossType = ChildType = TabType = WindowType = ScreenType = object
BadLineType = SequenceMap = KeyActionType = AliasMap = object
AddressFamily = PopenType = Socket = StartupCtx = object
SessionTab = SessionType = LayoutType = SpecialWindowInstance = object
MarkType = RemoteCommandType = object
KeyEventType = ImageManagerType = KittyCommonOpts = HandlerType = object
GRT_t = GRT_a = GRT_d = GRT_f = GRT_m = GRT_o = GRT_C = object
ScreenSize = KittensKeyActionType = MouseEvent = MouseButton = AbstractEventLoop = object
TermManagerType = LoopType = Debug = GraphicsCommandType = object
ReadableBuffer = WriteableBuffer = bytearray

CompletedProcess = tuple
TypedDict = dict
EdgeLiteral = str
UnderlineLiteral = str
PowerlineStyle = str
MatchType = str
Protocol = object
OptionsProtocol = object
NotRequired = tuple
CoreTextFont = FontConfigPattern = dict
